import datetime
import re
from typing import Iterator, List, Optional, Tuple

import attr

from .enums import ChangeLogTypeEnum, FormatTypeEnum


BREAKING_CHANGE_IN_BODY = "BREAKING CHANGE:"
BREAKING_CHANGE_IN_COMMIT_TYPE = "!"

COMMIT_TYPE_FEATURE = "feat"
COMMIT_TYPE_FIX = "fix"

COMMIT_TYPE_SUBJECT_RE = re.compile(
    r"^(?P<commit_type>[^\:]+)\: (?P<description>.+)$"
)
COMMIT_TYPE_SCOPE_RE = re.compile(
    r"^(?P<commit_type>[^\(]+)\((?P<scope>[^\)]+)\)$"
)

ISSUE_RE = re.compile(r"^Issue: (?P<issue>.+)$", re.M)


@attr.dataclass(frozen=True, slots=True)
class ConventionalCommit:
    raw_commit_type: str
    description: str

    body: Optional[str] = None

    @property
    def clean_commit_type(self) -> str:
        commit_type = self.raw_commit_type
        if commit_type[-1] == BREAKING_CHANGE_IN_COMMIT_TYPE:
            return commit_type[:-1]
        return commit_type

    @property
    def commit_type(self) -> str:
        commit_type = self.clean_commit_type

        maybe_matched = COMMIT_TYPE_SCOPE_RE.match(commit_type)
        if maybe_matched is not None:
            return maybe_matched.groupdict()["commit_type"]

        return commit_type

    def format(self, format_type: FormatTypeEnum) -> str:  # noqa: A003
        prefix = ""
        if self.scope:
            prefix = f"({bold(self.scope)}) "

        if self.is_breaking_change:
            prefix = f"{bold(BREAKING_CHANGE_IN_BODY)} {prefix}"

        issues = self.issues
        if issues:
            prefix = f'[{", ".join(issues)}] {prefix}'

        return f"{prefix}{self.description}"

    @classmethod
    def from_git_commit(cls, git_commit: str) -> "ConventionalCommit":
        subject, *body = git_commit.splitlines()

        maybe_matched = COMMIT_TYPE_SUBJECT_RE.match(subject)
        if maybe_matched is not None:
            matched_dict = maybe_matched.groupdict()
            return cls(
                raw_commit_type=matched_dict["commit_type"],
                description=matched_dict["description"],
                body="\n".join(body[1:]) if body else None,
            )

        raise ValueError("Unable to parse git commit as conventional commit")

    @property
    def is_breaking_change(self) -> bool:
        return self.raw_commit_type[-1] == BREAKING_CHANGE_IN_COMMIT_TYPE or (
            self.body is not None and BREAKING_CHANGE_IN_BODY in self.body
        )

    @property
    def issues(self) -> Tuple[str, ...]:
        if self.body is None:
            return ()
        return tuple(item.strip() for item in ISSUE_RE.findall(self.body))

    @property
    def scope(self) -> Optional[str]:
        commit_type = self.clean_commit_type

        maybe_matched = COMMIT_TYPE_SCOPE_RE.match(commit_type)
        if maybe_matched is not None:
            return maybe_matched.groupdict()["scope"]

        return None


@attr.dataclass(frozen=True, slots=True)
class ChangeLog:
    commits: Tuple[ConventionalCommit, ...]

    feature_commits: Tuple[ConventionalCommit, ...] = attr.ib(init=False)
    fix_commits: Tuple[ConventionalCommit, ...] = attr.ib(init=False)
    other_commits: Tuple[ConventionalCommit, ...] = attr.ib(init=False)

    def __attrs_post_init__(self) -> None:
        feature_commits: List[ConventionalCommit] = []
        fix_commits: List[ConventionalCommit] = []
        other_commits: List[ConventionalCommit] = []

        for commit in self.commits:
            if commit.commit_type == COMMIT_TYPE_FEATURE:
                feature_commits.append(commit)
            elif commit.commit_type == COMMIT_TYPE_FIX:
                fix_commits.append(commit)
            else:
                other_commits.append(commit)

        object.__setattr__(self, "feature_commits", tuple(feature_commits))
        object.__setattr__(self, "fix_commits", tuple(fix_commits))
        object.__setattr__(self, "other_commits", tuple(other_commits))

    def format(  # noqa: A003
        self,
        changelog_type: ChangeLogTypeEnum,
        format_type: FormatTypeEnum,
        *,
        is_pre_release: bool = False,
    ) -> str:
        is_git_commit = changelog_type == ChangeLogTypeEnum.git_commit
        is_rst = format_type == FormatTypeEnum.rst

        def format_block(
            label: str, commits: Tuple[ConventionalCommit, ...]
        ) -> Optional[str]:
            if not commits:
                return None

            header: str
            if is_rst:
                header = bold(label)
            elif is_git_commit:
                header = markdown_h2(label, git_safe=True)
            elif is_pre_release:
                header = markdown_h3(label)
            else:
                header = markdown_h2(label)

            breaking_items = format_commits(
                item for item in commits if item.is_breaking_change
            )
            regular_items = format_commits(
                item for item in commits if not item.is_breaking_change
            )
            items = "\n".join(
                item for item in (breaking_items, regular_items) if item
            )

            return "\n\n".join((header, items))

        def format_commits(commits: Iterator[ConventionalCommit]) -> str:
            return "\n".join(
                f"- {item.format(format_type)}" for item in commits
            )

        features = format_block("Features:", self.feature_commits)
        fixes = format_block("Fixes:", self.fix_commits)
        others = format_block("Other:", self.other_commits)

        return "\n\n".join(item for item in (features, fixes, others) if item)

    @classmethod
    def from_git_commits(cls, git_commits: Tuple[str, ...]) -> "ChangeLog":
        return cls(
            commits=tuple(
                ConventionalCommit.from_git_commit(item)
                for item in reversed(git_commits)
            )
        )

    @property
    def has_breaking_change(self) -> bool:
        return any(item for item in self.commits if item.is_breaking_change)

    @property
    def has_minor_change(self) -> bool:
        return len(self.feature_commits) > 0

    @property
    def has_micro_change(self) -> bool:
        return len(self.feature_commits) != len(self.commits)


def bold(value: str) -> str:
    return f"**{value}**"


def in_development_header(version: str, format_type: FormatTypeEnum) -> str:
    content = f"{version} (In Development)"
    return (
        rst_h1(content)
        if format_type == FormatTypeEnum.rst
        else markdown_h1(content)
    )


def markdown_h1(value: str, *, git_safe: bool = False) -> str:
    return rst_h1(value) if git_safe else markdown_header(value, level=1)


def markdown_h2(value: str, *, git_safe: bool = False) -> str:
    return rst_h2(value) if git_safe else markdown_header(value, level=2)


def markdown_h3(value: str) -> str:
    return markdown_header(value, level=3)


def markdown_header(value: str, *, level: int) -> str:
    return f'{"#" * level} {value}'


def rst_h1(value: str) -> str:
    return rst_header(value, symbol="=")


def rst_h2(value: str) -> str:
    return rst_header(value, symbol="-")


def rst_header(value: str, *, symbol: str) -> str:
    return f"{value}\n{symbol * len(value)}"


def ul_li(value: str) -> str:
    return f"- {value}"


def version_header(
    version: str,
    format_type: FormatTypeEnum,
    *,
    include_date: bool,
    is_pre_release: bool,
) -> str:
    content = version
    if include_date:
        now = datetime.datetime.utcnow()
        content = f"{version} ({now.date().isoformat()})"

    is_rst = format_type == FormatTypeEnum.rst
    functions = (
        # Release -> (Markdown, RST)
        (markdown_h1, rst_h1),
        # Pre-release -> (Markdown, RST)
        (markdown_h2, rst_h2),
    )

    return functions[int(is_pre_release)][int(is_rst)](content)



import subprocess
from pathlib import Path
from typing import Set, Tuple

import toml

from .output import diff, echo_message
from ..changelog import ChangeLog, in_development_header, version_header
from ..configs import find_changelog_file, ProjectConfig
from ..constants import (
    CHANGELOG_UPPER,
    FILE_PACKAGE_JSON,
    FILE_PACKAGE_LOCK_JSON,
    FILE_PYPROJECT_TOML,
    FILE_YARN_LOCK,
)
from ..enums import ChangeLogTypeEnum, FormatTypeEnum, ProjectTypeEnum
from ..exceptions import ConfigError
from ..versions import Version


def find_changelog_path(config: ProjectConfig) -> Path:
    path = config.path

    if config.changelog_format_type_file == FormatTypeEnum.markdown:
        default_file = f"{CHANGELOG_UPPER}.md"
        maybe_changelog_path = find_changelog_file(path, "*.md")
    else:
        default_file = f"{CHANGELOG_UPPER}.rst"
        maybe_changelog_path = find_changelog_file(path, "*.rst")

    if maybe_changelog_path:
        return maybe_changelog_path

    return path / default_file


def guess_version_files(config: ProjectConfig) -> Tuple[str, ...]:
    if config.project_type == ProjectTypeEnum.javascript:
        return (FILE_PACKAGE_JSON,)

    path = config.path
    version_files = []

    maybe_pyproject_toml_path = path / FILE_PYPROJECT_TOML
    if maybe_pyproject_toml_path.exists():
        version_files.append(FILE_PYPROJECT_TOML)

        project_name = (
            toml.loads(maybe_pyproject_toml_path.read_text())
            .get("tool", {})
            .get("poetry", {})
            .get("name")
        )

        if project_name:
            for package in (".", "./src"):
                package_path = path / package

                if (package_path / project_name / "__init__.py").exists():
                    version_files.append(
                        f"{package}/{project_name}/__init__.py"
                    )

                if (package_path / project_name / "__version__.py").exists():
                    version_files.append(
                        f"{package}/{project_name}/__version__.py"
                    )

                if (package_path / f"{project_name}.py").exists():
                    version_files.append(f"{package}/{project_name}.py")

    return tuple(version_files)


def run_post_bump_hook(
    config: ProjectConfig, *, is_dry_run: bool = False
) -> None:
    """Run post-bump hook after version files already updated.

    Read command to execute from project config, or if it is not specified
    run ``npm install`` or ``yarn install`` for JavaScript projects due to
    presence of lock files in project path.
    """
    path = config.path
    cmd = config.post_bump_hook

    if cmd is None and config.project_type == ProjectTypeEnum.javascript:
        if (path / FILE_PACKAGE_LOCK_JSON).exists():
            cmd = "npm install"
        elif (path / FILE_YARN_LOCK).exists():
            cmd = "yarn install"

    if cmd is None:
        return None

    echo_message(f"Running post-bump hook: {cmd}", is_dry_run=is_dry_run)
    if is_dry_run:
        return None

    subprocess.check_call(cmd, cwd=path, shell=True)


def update_changelog_file(
    config: ProjectConfig,
    next_version: Version,
    changelog: ChangeLog,
    *,
    is_dry_run: bool = False,
) -> None:
    """Update changelog file with new version.

    In most cases it just prepend new changelog on top of the file, but if next
    version is pre-release, do little trickery instead to include
    In Development header as well.
    """
    changelog_path = find_changelog_path(config)
    next_version_str = next_version.format(config=config)

    echo_message(
        f"Adding {next_version_str} release notes to {changelog_path.name} "
        "file",
        is_dry_run=is_dry_run,
    )
    if is_dry_run:
        return

    dev_header = in_development_header(
        next_version.version.format(), config.changelog_format_type_file
    )
    changelog_content = changelog.format(
        ChangeLogTypeEnum.changelog_file,
        config.changelog_format_type_file,
        is_pre_release=next_version.pre_release is not None,
    )

    if next_version.pre_release is None:
        next_version_changelog = "\n\n".join(
            (
                version_header(
                    next_version_str,
                    config.changelog_format_type_file,
                    is_pre_release=False,
                    include_date=config.changelog_file_include_date,
                ),
                changelog_content,
            )
        )
    else:
        next_version_changelog = "\n\n".join(
            (
                dev_header,
                version_header(
                    next_version_str,
                    config.changelog_format_type_file,
                    is_pre_release=True,
                    include_date=config.changelog_file_include_date,
                ),
                changelog_content,
            )
        )

    next_version_changelog = f"{next_version_changelog}\n\n"

    if not changelog_path.exists():
        changelog_path.write_text(f"{next_version_changelog.strip()}\n")
    else:
        if not update_file(
            changelog_path, f"{dev_header}\n\n", next_version_changelog
        ):
            changelog_path.write_text(
                "".join((next_version_changelog, changelog_path.read_text()))
            )


def update_file(
    path: Path,
    current_content: str,
    next_content: str,
) -> bool:
    """
    Attempt to read file if it exists and update current content to next
    content.

    If file does not exists, or current content not found - return False,
    otherwise return True.
    """
    if not path.exists():
        return False

    # TODO: Do not read whole file output for make replacement
    content = path.read_text()
    if current_content not in content:
        return False

    next_content = content.replace(current_content, next_content)
    echo_message(diff(content, next_content), is_dry_run=False)

    path.write_text(next_content)
    return True


def update_version_files(
    config: ProjectConfig,
    current_version: Version,
    next_version: Version,
    *,
    is_dry_run: bool = False,
) -> bool:
    """Update all project version files.

    If they not specified in config, attempt to guess them automatically.
    """
    version_files = config.version_files
    if not version_files:
        version_files = guess_version_files(config)

    path = config.path
    current_version_str = current_version.format(config=config)
    next_version_str = next_version.format(config=config)

    updated: Set[bool] = set()

    for item in version_files:
        if item.startswith("..") or item.startswith("/"):
            raise ConfigError(
                "Version file outside of project directory is forbidden: "
                f"{item}"
            )

        echo_message(
            f"Updating version in {item}: {current_version_str} -> "
            f"{next_version_str}",
            is_dry_run=is_dry_run,
        )
        if is_dry_run:
            continue

        updated.add(
            update_file(
                path.joinpath(item),
                current_version_str,
                next_version_str,
            )
        )

    return len(updated) == 1 and updated.pop() is True



import re
from typing import List


CO_AUTHORED_BY = "Co-authored-by: "
COMMIT_SUBJECT_WITH_PR_RE = re.compile(r"^(?P<subject>.+) \(\#\d+\)$")


def clean_body(body: List[str]) -> str:
    cleaned = "\n".join(
        item for item in body if not item.startswith(CO_AUTHORED_BY)
    )
    if cleaned[-1:] == "\n":
        return cleaned
    return f"{cleaned}\n"


def clean_commit_subject(value: str) -> str:
    return COMMIT_SUBJECT_WITH_PR_RE.sub(r"\1", value)


def clean_tag_ref(value: str) -> str:
    if value[:10] == "refs/tags/":
        return value[10:]
    return value



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['badabump', 'badabump.ci', 'badabump.cli', 'badabump.versions']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=20.2.0,<21.0.0', 'toml>=0.10.1,<0.11.0']

entry_points = \
{'console_scripts': ['badabump = badabump.cli.app:main',
                     'badabump-ci = badabump.cli.ci_app:main']}

setup_kwargs = {
    'name': 'badabump',
    'version': '20.1.0b0',
    'description': 'Manage changelog and bump project version number using conventional commits from latest git tag. Support Python, JavaScript projects. Understand CalVer, SemVer version schemas. Designed to run at GitHub Actions.',
    'long_description': '# badabump\n\n[![CalVer](https://img.shields.io/badge/calver-YY.MINOR.MICRO-22bfda)](https://calver.org)\n[![CI Workflow](https://github.com/playpauseandstop/badabump/workflows/ci/badge.svg)](https://github.com/playpauseandstop/badabump/actions?query=workflow%3A%22ci%22)\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)](https://pre-commit.com)\n[![Commitizen friendly](https://img.shields.io/badge/commitizen-friendly-brightgreen.svg)](https://github.com/commitizen-tools/commitizen#integrating-with-pre-commit)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Latest Version](https://img.shields.io/pypi/v/badabump.svg)](https://pypi.org/project/badabump/)\n[![Python Versions](https://img.shields.io/pypi/pyversions/badabump.svg)](https://pypi.org/project/badabump/)\n[![BSD License](https://img.shields.io/pypi/l/badabump.svg)](https://github.com/playpauseandstop/badabump/blob/master/LICENSE)\n[![Coverage](https://coveralls.io/repos/playpauseandstop/badabump/badge.svg?branch=master&service=github)](https://coveralls.io/github/playpauseandstop/badabump)\n\nManage changelog and bump project version number using conventional commits from latest git tag. Support Python, JavaScript projects. Understand CalVer, SemVer version schemas. Designed to run at GitHub Actions.\n',
    'author': 'Igor Davydenko',
    'author_email': 'iam@igordavydenko.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://igordavydenko.com/projects/#badabump',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import datetime
from typing import Optional

import attr

from .exceptions import VersionError, VersionParseError
from .formatting import format_version
from .parsing import parse_version
from ..annotations import DictStrAny, DictStrStr
from ..configs import UpdateConfig
from ..constants import DEFAULT_VERSION_SCHEMA


SHORT_YEAR_START = 2000

SCHEMA_PARTS_FORMATTING = {
    "YYYY": "{year}",
    "YY": "{short_year}",
    "0Y": "{short_year:02d}",
    "MM": "{month}",
    "0M": "{month:02d}",
    "WW": "{week}",
    "0W": "{week:02d}",
    "DD": "{day}",
    "0D": "{day:02d}",
    "MINOR": "{minor}",
    "MICRO": "{micro}",
}
SCHEMA_PARTS_PARSING = {
    "YYYY": r"(?P<year>\d{4})",
    "YY": r"(?P<short_year>\d{2})",
    "0Y": r"(?P<short_year>\d{2})",
    "MM": r"(?P<month>\d{1,2})",
    "0M": r"(?P<month>\d{2})",
    "WW": r"(?P<week>\d{1,2})",
    "0W": r"(?P<week>\d{2})",
    "DD": r"(?P<day>\d{1,2})",
    "0D": r"(?P<day>\d{2})",
    "MINOR": r"(?P<minor>\d+)",
    "MICRO": r"(?P<micro>\d+)",
}


@attr.dataclass(frozen=True, slots=True)
class CalVer:
    year: int
    month: Optional[int] = None
    week: Optional[int] = None
    day: Optional[int] = None
    minor: Optional[int] = None
    micro: Optional[int] = None

    schema: str = DEFAULT_VERSION_SCHEMA

    def format(self) -> str:  # noqa: A003
        return format_version(
            self.schema, SCHEMA_PARTS_FORMATTING, self.get_format_context()
        )

    @classmethod
    def from_parsed_dict(cls, parsed: DictStrStr, *, schema: str) -> "CalVer":
        maybe_year = guess_year(parsed.get("year"), parsed.get("short_year"))
        if maybe_year is None:
            raise VersionError("CalVer version should contains at least year")

        return cls(
            year=maybe_year,
            month=int_or_none(parsed.get("month")),
            week=int_or_none(parsed.get("week")),
            day=int_or_none(parsed.get("day")),
            minor=int_or_none(parsed.get("minor")),
            micro=int_or_none(parsed.get("micro")),
            schema=schema,
        )

    def get_format_context(self) -> DictStrAny:
        return {**attr.asdict(self), "short_year": self.short_year}

    @classmethod
    def parse(cls, value: str, *, schema: str) -> "CalVer":
        maybe_parsed = parse_version(schema, SCHEMA_PARTS_PARSING, value)
        if maybe_parsed:
            return cls.from_parsed_dict(maybe_parsed, schema=schema)
        raise VersionParseError(schema, value)

    @property
    def short_year(self) -> Optional[int]:
        if self.year is None:
            return None
        if self.year < SHORT_YEAR_START:
            return None
        return self.year - SHORT_YEAR_START

    def update(self, config: UpdateConfig) -> "CalVer":
        utcnow = datetime.datetime.utcnow()

        next_minor = self.minor
        next_micro = self.micro

        if not config.is_pre_release:
            if next_minor is not None and (
                config.is_breaking_change or config.is_minor_change
            ):
                next_minor += 1
                next_micro = 0 if next_micro is not None else None

            if next_micro is not None and config.is_micro_change:
                next_micro += 1

        return attr.evolve(
            self,
            year=utcnow.year if self.year else None,
            month=utcnow.month if self.month else None,
            week=get_week(utcnow) if self.week else None,
            day=utcnow.day if self.day else None,
            minor=next_minor,
            micro=next_micro,
        )


def get_week(value: datetime.datetime) -> int:
    return value.isocalendar()[1]


def guess_year(
    year: Optional[str], short_year: Optional[str]
) -> Optional[int]:
    if year is not None:
        return int(year)
    if short_year is not None:
        return SHORT_YEAR_START + int(short_year)
    return None


def int_or_none(value: Optional[str]) -> Optional[int]:
    if value is not None:
        return int(value)
    return None



import argparse
from pathlib import Path


def add_path_argument(parser: argparse.ArgumentParser) -> argparse.Action:
    return parser.add_argument(
        "-C",
        "--path",
        default=Path.cwd(),
        help="Directory with project. By default: current working directory",
        type=Path,
    )
